"""this module provides _TypeChecker."""
import datetime
from types import GenericAlias
from typing import Any, Type, Union, _SpecialForm, get_args, get_origin

from fastconfig.exception import UnexpectedValueError

DATE_TYPES = [datetime.datetime, datetime.date, datetime.time]


class _TypeChecker:
    def __call__(
        self, key: str, value: Any, typeinfo: Union[type, GenericAlias, _SpecialForm]
    ) -> Any:
        self.value = value
        if not self.check(key, value, typeinfo):
            raise UnexpectedValueError(
                f"{key}: {value} is not valid type. must be of type {typeinfo}"
            )
        return self.value

    def check(
        self, key: str, value: Any, typeinfo: Union[type, GenericAlias, _SpecialForm]
    ) -> bool:
        if typeinfo == Any:
            return True
        outside = get_origin(typeinfo)
        if outside is None:
            if any(typeinfo is ty for ty in DATE_TYPES):
                return self.check_datetime(value, typeinfo)  # type: ignore
            return isinstance(value, typeinfo)  # type: ignore
        return self.check_origin(key, value, typeinfo)

    def check_origin(
        self,
        key: str,
        value: Any,
        typeinfo: Union[type, GenericAlias, _SpecialForm],
    ) -> bool:
        outside = get_origin(typeinfo)
        types = get_args(typeinfo)
        if outside == Union:
            return any(self.check(key, value, ty) for ty in types)
        elif outside == dict:
            k, v = types
            if not isinstance(value, dict):
                return False
            return all(self.check(key, key, k) for key in value.keys()) and all(
                self.check(key, val, v) for val in value.values()
            )
        elif outside == list:
            content = types[0]
            if not isinstance(value, list):
                return False
            return all(self.check(key, val, content) for val in value)
        else:
            raise UnexpectedValueError(f"{typeinfo} is not supported")

    def check_datetime(
        self,
        value: Union[str, datetime.datetime, datetime.date, datetime.time],
        typeinfo: Union[
            Type[datetime.datetime], Type[datetime.date], Type[datetime.time]
        ],
    ) -> Any:
        if type(value) is typeinfo:
            return True

        if isinstance(value, str):
            if typeinfo is datetime.time:
                return False
            try:
                """
                TODO:
                ISO 8601 and RFC 3339 is not compatible
                However, there are few use cases for writing the specified date and time to the setting value,
                so I left it as it is for now.
                """
                value = datetime.datetime.fromisoformat(value)
            except ValueError:
                return False

        if not isinstance(value, datetime.datetime):
            return False

        if typeinfo is datetime.date:
            value = value.date()
        elif typeinfo is datetime.time:
            value = value.time()

        self.value = value
        return True



"""this module provides Validator."""
from dataclasses import MISSING, Field
from typing import Any, List, Optional, Union

from fastconfig.exception import MissingRequiredElementError
from fastconfig.internals.type_checker import _TypeChecker


def _extract(setting: dict[str, Any], section: Union[str, List[str]]) -> Optional[Any]:
    if isinstance(section, str) and section in setting:
        return setting[section]
    elif isinstance(section, List):
        for element in section:
            if not (isinstance(setting, dict) and element in setting):
                return None
            setting = setting[element]
        return setting
    return None


class DEFAULT_VALUE:
    """This class is the dummy object for meaning `use default value`."""

    pass


class _Validator:
    def __init__(self, setting: dict[str, Any]) -> None:
        self.setting: dict[str, Any] = setting
        self.checker: _TypeChecker = _TypeChecker()

    def __call__(self, key: str, f: Field, build: bool = True) -> Any:
        metadata: dict[str, Any] = dict(f.metadata) if hasattr(f, "metadata") else {}
        separator = metadata["separator"] if "separator" in metadata else "."
        setting_key: Union[str, List[str]] = (
            metadata["key"].split(separator) if "key" in metadata else key
        )
        value: Any = _extract(self.setting, setting_key)

        if value is None:
            if (
                isinstance(f.default, type(MISSING))
                and isinstance(f.default_factory, type(MISSING))
                and build
            ):
                # TODO: check default_factry
                raise MissingRequiredElementError(f"key: {key} is not found")
            return DEFAULT_VALUE()

        value = self.checker(key, value, f.type)
        return value



"""this module provides Exceptions in FastConfig."""


class FastConfigError(Exception):
    """super class for all errors returned by FastConfig."""

    ...


class InvalidConfigError(FastConfigError):
    """exception when providing files or classes in unsupported formats."""

    ...


class MissingRequiredElementError(FastConfigError):
    """exception when providing `field` has no default value and the corresponding value could not be read from the file."""

    ...


class UnexpectedValueError(FastConfigError):
    """exception whe providing failed type-checking."""

    ...



"""this module provides FastConfig class."""
import os
import sys
from dataclasses import MISSING, asdict, dataclass, field
from pathlib import Path
from typing import Any, Mapping, Optional, Type, TypeVar, Union

from fastconfig.exception import InvalidConfigError
from fastconfig.internals.loader import _FileLoader
from fastconfig.internals.validator import DEFAULT_VALUE, _Validator

_T = TypeVar("_T")
_Self = TypeVar("_Self", bound="FastConfig")


@dataclass
class FastConfig:
    """this class provides the way to build and update instance from `toml` or `json` format file."""

    @classmethod
    def build(
        cls: Type[_Self], path: Union[str, Path], config: Optional[_Self] = None
    ) -> _Self:
        """
        Read file from path and create/update instance.

        Args:
            path: Union[str, Path]
                a file path to read a config
            config: Optional[_Self]
                an instance inheriting from FastConfig (if updating)
        Returns:
            _Self: an instance inheriting from FastConfig
        """
        if config is None:
            return _FastConfigBuilder.build(path, cls)
        else:
            return _FastConfigBuilder.build(path, config)

    def to_dict(self, use_key: bool = False) -> dict[str, Any]:
        """
        Convert from an instance to dict.

        Args:
            use_key: bool
                Whether to convert to dictionary by class instance name
        Returns:
            dict[str, Any]
        """
        if use_key:
            return asdict(self)

        dic: dict[str, Any] = {}
        for key, f in self.__dataclass_fields__.items():
            metadata: dict[str, Any] = (
                dict(f.metadata) if hasattr(f, "metadata") else {}
            )
            separator = metadata["separator"] if "separator" in metadata else "."
            setting_key: list[str] = (
                metadata["key"].split(separator) if "key" in metadata else [key]
            )
            inside: dict[str, Any] = dic
            for nest_field in setting_key[:-1]:
                if nest_field not in dic:
                    inside[nest_field] = {}
                inside = inside[nest_field]
            inside[setting_key[-1]] = getattr(self, key)
        return dic


if sys.version_info >= (3, 10):

    def fc_field(
        key: Optional[str | int] = None,
        separator: str = ".",
        default: _T = MISSING,
        default_factory: Type[_T] = MISSING,
        init: bool = True,
        repr: bool = True,
        hash: Optional[bool] = None,
        compare: bool = True,
        metadata: Mapping[Any, Any] | None = None,
        kw_only=MISSING,
    ) -> _T:
        """
        Return `dataclass::field`.

        wrapper of `dataclass::field` for FastConfig definition.
        same as `field` except for passing `key` and `separator`.

        Args:
            key: Optional[str | int]
                the name of the key you want to read, divided by `separator` when retrieving data
            separator: str
                Separator for splitting key values, default is `.`

        Returns:
            _T
        """
        options: dict[str, Any] = {
            "default": default,
            "default_factory": default_factory,
            "init": init,
            "repr": repr,
            "hash": hash,
            "compare": compare,
            "metadata": metadata if metadata is not None else {},
            "kw_only": kw_only,
        }

        if key is not None:
            options["metadata"]["key"] = key
        options["metadata"]["separator"] = separator
        return field(**options)

else:

    def fc_field(
        key: Optional[Union[str, int]] = None,
        separator: str = ".",
        default: _T = MISSING,
        default_factory: Type[_T] = MISSING,
        init: bool = True,
        repr: bool = True,
        hash: Optional[bool] = None,
        compare: bool = True,
        metadata: Optional[Mapping[Any, Any]] = None,
    ) -> _T:
        """
        Return `dataclass::field`.

        wrapper of `dataclass::field` for FastConfig definition.
        same as `field` except for passing `key` and `separator`.

        Args:
            key: Optional[str | int]
                the name of the key you want to read, divided by `separator` when retrieving data
            separator: str
                Separator for splitting key values, default is `.`

        Returns:
            _T
        """
        options: dict[str, Any] = {
            "default": default,
            "default_factory": default_factory,
            "init": init,
            "repr": repr,
            "hash": hash,
            "compare": compare,
            "metadata": metadata if metadata is not None else {},
        }

        if key is not None:
            options["metadata"]["key"] = key
        options["metadata"]["separator"] = separator
        return field(**options)


class _FastConfigBuilder:
    @classmethod
    def build(cls, path: Union[str, Path], config: Union[_Self, Type[_Self]]) -> _Self:
        if isinstance(path, Path):
            path = str(path)

        if not os.path.exists(path):
            raise FileNotFoundError(f"{path} is not found")

        loader: _FileLoader = _FileLoader()
        data: dict[str, Any] = loader(path)
        if not isinstance(config, type) and isinstance(config, FastConfig):
            return cls._update(config, data)
        elif isinstance(config, type) and issubclass(config, FastConfig):
            return cls._make(config, data)
        else:
            raise InvalidConfigError(
                "must be of type FastConfig or an instance of FastConfig"
            )

    @classmethod
    def _make(
        cls,
        config: Type[_Self],
        setting: dict[str, Any],
    ) -> _Self:
        # check metadata and type hint
        args: dict[str, Any] = {}
        checker: _Validator = _Validator(setting)
        for key, f in config.__dataclass_fields__.items():
            value = checker(key, f)
            if not isinstance(value, DEFAULT_VALUE):
                args[key] = value
        return config(**args)

    @classmethod
    def _update(cls, config: _Self, setting: dict[str, Any]) -> _Self:
        checker: _Validator = _Validator(setting)
        for key, f in config.__dataclass_fields__.items():
            value = checker(key, f, build=False)
            if not isinstance(value, DEFAULT_VALUE):
                setattr(config, key, value)
        return config



"""this module provides `_FileLoader`."""
import json
from typing import Any

import toml

from fastconfig.exception import InvalidConfigError


class _FileLoader:
    def __call__(self, path: str) -> dict[str, Any]:
        if path.endswith(".json"):
            return self.load_json(path)
        elif path.endswith(".toml"):
            return self.load_toml(path)
        raise InvalidConfigError("FastConfig only supports json and toml formats now")

    def load_json(self, path: str) -> dict[str, Any]:
        try:
            with open(path, "r") as f:
                config: Any = json.load(f)
            if not isinstance(config, dict):
                config = {"content": config}
            return config
        except json.decoder.JSONDecodeError as e:
            raise InvalidConfigError(str(e))

    def load_toml(self, path: str) -> dict[str, Any]:
        try:
            with open(path, "r") as f:
                config = toml.load(f)
        except toml.decoder.TomlDecodeError as e:
            raise InvalidConfigError(str(e))
        return config



"""This package provides the methods to search files."""
import os
from pathlib import Path
from typing import List, Optional, Union

__version__ = "0.2.0"
PROJECT_ROOTS: List[str] = [".hg", ".git"]
DEPTH: int = 10


def is_project_root(path: Union[str, Path]) -> bool:
    """
    Check the given path is the project root directory or not.

    This method determines the project root by whether the version control tool directory exists.

    Args:
        path (Union[str, Path]):
            The path to check whether it is the project root or not

    Returns:
        bool: The result of the project root or not
    """
    if isinstance(path, str):
        path = Path(path)

    if path.is_file():
        path = path.parent

    for candidate in PROJECT_ROOTS:
        if path.joinpath(candidate).is_dir():
            return True
    return False


def find_project_root(path: Optional[Union[str, Path]] = None) -> Optional[Path]:
    """
    Return if the project root is found, or None if not.

    Args:
        path (Optional[Union[str, Path]]):
            A path string or Path object to start searching, If nothing is passed, start in the current directory

    Returns:
        Optional[Path]: the project root path, or None if the project root is not found
    """
    if path is None:
        path = os.getcwd()
    cnt: int = 0

    candidate: Path = Path(path) if isinstance(path, str) else path
    if is_project_root(candidate):
        return candidate

    while cnt < DEPTH:
        candidate = candidate.parent
        if is_project_root(candidate):
            return candidate

        if candidate.parent == candidate:
            return None

        cnt += 1
    return None


def search(
    target: Union[str, Path],
    path: Optional[Union[str, Path]] = None,
    end_up_the_project_root: bool = True,
) -> Optional[Path]:
    """
    Recursively searches for files with the name of the target and returns the result.

    Args:
        target (Union[str, Path]):
            Search target filename, and directory names are ignored.

        path (Optional[Union[str, Path]]):
            A path string or Path object to start searching, If nothing is passed, start in the current directory.

        end_up_the_project_root (bool):
            Whether or not to search the directory where the version control tool exists

    Returns:
        Optional[Path]: a path of the target file, or None if the target file is not found
    """
    if isinstance(target, str):
        target = Path(target)

    target_name: str = target.name
    if path is None:
        path = os.getcwd()

    cnt: int = 0
    candidate: Path = Path(os.path.join(path, target_name))
    if candidate.exists():
        return candidate
    if is_project_root(candidate):
        return None

    while cnt < DEPTH:
        candidate = candidate.parent.parent.joinpath(target_name)
        if candidate.exists():
            return candidate

        if candidate.parent.parent.joinpath(target_name) == candidate or (
            end_up_the_project_root and is_project_root(candidate)
        ):
            return None

        cnt += 1
    return None
